#include "kalman.h"

/*******************************************************
* Kalman filter PUMA 3DOF Robotic Arm
********************************************/
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
//my matrix library, you can use your own favorite matrix library
#define TIMESTEP 0.05

int matrixMul(a[6][6], b[6][1], c[6][1]){
	int i,j;
	for(i = 0; i < 6; i++){
		int sum = 0;
		for(j = 0; j < 6; j++)
			sum += a[i][j] * b[j][0];
		c[i][0] = sum;
	}
}

float timeIndex,m_a1,m_a2,m_a3,c_a1,c_a2,c_a3,tau1,tau2,tau3;
float a1,a2,a3,a1d,a2d,a3d,a1dd,a2dd,a3dd;
float new_a1d, new_a2d, new_a3d;
float TIME=0;
float A[6][6];
float B[6][3];
float C[6][6] = {1, 0, 0, 0, 0, 0, 
				0, 1, 0, 0, 0, 0,
				0, 0, 1, 0, 0, 0,
				0, 0, 0, 1, 0, 0,
				0, 0, 0, 0, 1, 0,
				0, 0, 0, 0, 0, 1};
float Q[6][6] = {1, 0, 0, 0, 0, 0, 
				0, 1, 0, 0, 0, 0,
				0, 0, 1, 0, 0, 0,
				0, 0, 0, 1, 0, 0,
				0, 0, 0, 0, 1, 0,
				0, 0, 0, 0, 0, 1};
float R[6][6] = {1, 0, 0, 0, 0, 0, 
				0, 1, 0, 0, 0, 0,
				0, 0, 1, 0, 0, 0,
				0, 0, 0, 1, 0, 0,
				0, 0, 0, 0, 1, 0,
				0, 0, 0, 0, 0, 1};
float y[6][1];
float K[6][6];
float x[6][1];
float state[6][1];
float action[3][1];
float lastState[6][1];
float P[6][6] = {1, 0, 0, 0, 0, 0, 
				0, 1, 0, 0, 0, 0,
				0, 0, 1, 0, 0, 0,
				0, 0, 0, 1, 0, 0,
				0, 0, 0, 0, 1, 0,
				0, 0, 0, 0, 0, 1};
float p[6][6] = {1, 0, 0, 0, 0, 0, 
				0, 1, 0, 0, 0, 0,
				0, 0, 1, 0, 0, 0,
				0, 0, 0, 1, 0, 0,
				0, 0, 0, 0, 1, 0,
				0, 0, 0, 0, 0, 1};
float measurement[6][1];

void initKalman(){
	float a[6][6]={ 1.004,    0.0001,  0.001,    0.0014,    0.0000,  -0.0003  ,
					0.000,    1.000,     -0.00,      0.0000,    0.0019,   0 ,
					0.0004,  0.0002,  1.002,    0.0003,    0.0001,   0.0015   ,
					0.2028,  0.0481,  0.0433,  0.7114,   -0.0166,  -0.1458   ,
					0.0080,  0.0021, -0.0020, -0.0224,    0.9289,   0.005    ,
					0.1895,  0.1009,  0.1101, -0.1602,    0.0621,   0.7404 };

	float b[6][3] = {0.0000,   0.0000 ,  0.0000,
					0.0000,   0.0000,  -0.0000,
					0.0000,  -0.0000,   0.0000 ,
					0.007,    -0.0000,   0.0005,
					0.0001,   0.0000,  -0.0000,
					0.0003,  -0.0000,   0.0008};

/*loads the A and B matrices from above*/
	for (int i = 0; i < 6; i++){
		for (int j = 0; j < 6; j++){
			A[i][j]=a[i][j];
		}
	}
	
	for (int i = 0; i < 6; i++){
		for (int j = 0; j < 3; j++){
			B[i][j]=b[i][j];
		}
	}

	/*initializes the state*/
	state[0][0]=0.1;
	state[1][0]=0.1;
	state[2][0]=0.1;
	state[3][0]=0.1;
	state[4][0]=0.1;
	state[5][0]=0.1;

	lastState=state;
}

void kalman(){
	int i, j = 0;
	lastState=state;
	state[0][0]=c_a1;
	state[1][0]=c_a2;
	state[2][0]=c_a3;
	state[3][0]=a1d;
	state[4][0]=a2d;
	state[5][0]=a3d;

	measurement[0][0]=m_a1;
	measurement[1][0]=m_a2;
	measurement[2][0]=m_a3;
	measurement[3][0]=a1d;
	measurement[4][0]=a2d;
	measurement[5][0]=a3d;

	action[0][0]=tau1;
	action[1][0]=tau2;
	action[2][0]=tau3;

	float temp1[6][6];
	float temp2[6][6];
	float temp3[6][6];
	float temp4[6][1];
/************ Prediction Equations*****************/
	for(i = 0; i < 6; i ++){
		
		for(j = 0; j < 6; j++)
	}
	x = A*lastState + B*action;
	p = A*P*A + Q;
/************ Update Equations **********/
K = p*C*pinv(C*p*C+R);

y=C*state;

state = x + K*(y-C*lastState);

P = (eye(6)  K*C)*p;

a1=state[0][0];
a2=state[1][0];
a3=state[2][0];
a1d=state[3][0];
a2d=state[4][0];
a3d=state[5][0];
}

/* This function is not used since I am using position to get velocity (i.e. differentiation).
* However I think that it is useful to include if you want velocity and position
* from acceleration you would use it */
void integrate(){
new_a1d = a1d + a1dd*TIMESTEP;
a1 += (new_a1d + a1d)*TIMESTEP/2;
a1d = new_a1d;
new_a2d = a2d + a2dd*TIMESTEP;
a2 += (new_a2d + a2d)*TIMESTEP/2;
a2d = new_a2d;
new_a3d = a3d + a3dd*TIMESTEP;
a3 += (new_a3d + a3d)*TIMESTEP/2;
a3d = new_a3d;
TIME+=TIMESTEP;
}

/*This gives me velocity from position*/
void differentiation(){
a1d=(state[0][0]-lastState[0][0])/TIMESTEP;
a2d=(state[1][0]-lastState[1][0])/TIMESTEP;
a3d=(state[2][0]-lastState[2][0])/TIMESTEP;
TIME+=TIMESTEP;
}

int callKalman(float *ti, float *ma1, float *ma2, float *ma3, float *ca1, 
		float *ca2, float *ca3, float *ta1, float *ta2, float *ta3){
	
	timeIndex = tm;
	m_a1 = ma1;
	m_a2 = ma2;
	m_a3 = ma3;
	c_a1 = ca1;
	c_a2 = ca2;
	c_a3 = ca3;
	tau1 = ta1;
	tau2 = ta2;
	tau3 = ta3;
	kalman();
	differentiation();
	//integrate();
	
	return 1;
}